import os
import scanpy as sc
import muon as mu
import anndata as ad

from anndata import AnnData
from muon import MuData

import numpy as np
import pandas as pd
import scipy
from scipy.sparse import csr_matrix
from scipy.sparse import issparse, csr
from sklearn.preprocessing import MaxAbsScaler

# from ._pca import select_pcs_features
from ._utils import reindex, gene_scores, TFIDF_LSI, clr_normalize_each_cell
from .adj_construction import create_adj, graph_construction

CHUNK_SIZE = 20000

def batch_scale(adata, use_rep='X', batch_key='batch', chunk_size=CHUNK_SIZE):
    """
    Batch-specific scale data

    Parameters
    ----------
    adata
        AnnData
    use_rep
        use '.X' or '.obsm'
    chunk_size
        chunk large data into small chunks

    """
    for b in adata.obs[batch_key].unique():
        idx = np.where(adata.obs[batch_key] == b)[0]
        if use_rep == 'X':
            scaler = MaxAbsScaler(copy=False).fit(adata.X[idx])
            for i in range(len(idx) // chunk_size + 1):
                adata.X[idx[i * chunk_size:(i + 1) * chunk_size]] = scaler.transform(
                    adata.X[idx[i * chunk_size:(i + 1) * chunk_size]])
        else:
            scaler = MaxAbsScaler(copy=False).fit(adata.obsm[use_rep][idx])
            for i in range(len(idx) // chunk_size + 1):
                adata.obsm[use_rep][idx[i * chunk_size:(i + 1) * chunk_size]] = scaler.transform(
                    adata.obsm[use_rep][idx[i * chunk_size:(i + 1) * chunk_size]])

    return adata


def preprocessing_rna(
        adata: AnnData,
        min_features: int = 600,
        min_cells: int = 3,
        target_sum: int = 10000,
        used_hvgs: bool = True,
        used_pca_graph: bool = True,
        rna_n_top_features=2000,  # or gene list
        n: int = 15,
        batch_key: str = 'batch',
        metric: str = 'euclidean',
        svd_solver: str = 'arpack',
        keep_mt: bool = False,
        backed: bool = False
):
    """
    Preprocessing single-cell RNA-seq data

    Parameters
    ----------
    adata
        An AnnData matrice of shape n_obs x n_vars. Rows correspond to cells and columns to genes.
    min_features
        Filtered out cells that are detected in less than n genes. Default: 600.
    min_cells
        Filtered out genes that are detected in less than n cells. Default: 3.
    target_sum
        After normalization, each cell has a total count equal to target_sum. If None, total count of each cell equal to the median of total counts for cells before normalization.
    rna_n_top_features
        Number of highly-variable genes to keep. Default: 2000.

    Return
    -------
    The AnnData object after preprocessing.
    """
    if min_features is None: min_features = 600
    if rna_n_top_features is None: rna_n_top_features = 2000
    if target_sum is None: target_sum = 10000

    # Preprocessing
    # if not issparse(adata.X):
    if type(adata.X) != csr.csr_matrix and (not backed) and (not adata.isbacked):
        adata.X = scipy.sparse.csr_matrix(adata.X)

    if not keep_mt:
        # Filtering out MT genes
        adata = adata[:, [gene for gene in adata.var_names
                          if not str(gene).startswith(tuple(['ERCC', 'MT-', 'mt-']))]]

    # Filtering cells
    sc.pp.filter_cells(adata, min_genes=min_features)

    # Filtering features
    sc.pp.filter_genes(adata, min_cells=min_cells)

    # Normalizing total per cell
    if adata.X.max() < 50:
        print('Warning: adata_rna.X may have already been normalized, do not normalize, please check.')
        adata.layers['norm_data'] = adata.X.copy()
    else:
        sc.pp.normalize_total(adata, target_sum=target_sum)
        # Log1p transforming
        sc.pp.log1p(adata)
        adata.layers['norm_data'] = adata.X.copy()

    # Finding variable features for RNA adata
    adata_hvg = adata.copy()
    if used_hvgs:
        if type(rna_n_top_features) == int:
            if rna_n_top_features > len(adata_hvg.var_names):
                rna_n_top_features = len(adata_hvg.var_names)
            if batch_key is not None:
                sc.pp.highly_variable_genes(adata_hvg, n_top_genes=rna_n_top_features, batch_key=batch_key)
            else:
                sc.pp.highly_variable_genes(adata_hvg, n_top_genes=rna_n_top_features)
                # sc.pp.highly_variable_genes(adata, min_mean=0.0125, max_mean=3, min_disp=0.5)
            adata_hvg = adata_hvg[:, adata_hvg.var.highly_variable].copy()
        elif type(rna_n_top_features) != int:
            if isinstance(rna_n_top_features, str):
                if os.path.isfile(rna_n_top_features):
                    rna_n_top_features = np.loadtxt(rna_n_top_features, dtype=str)
            adata_hvg = reindex(adata_hvg, rna_n_top_features)

    # scale data, clip values exceeding standard deviation 10.
    if adata_hvg.X.min() < 0:
        print('Warning: adata.X may have already been scaled, do not scale, please check.')
    else:
        sc.pp.scale(adata_hvg, max_value=10)
    # PCA
    sc.tl.pca(adata_hvg, svd_solver=svd_solver)
    # 返回norm_data
    adata_hvg.X = adata_hvg.layers['norm_data'].copy()

    if used_pca_graph:
        # use scanpy functions to do the graph construction
        sc.pp.neighbors(adata_hvg, n_neighbors=n, metric=metric, use_rep='X_pca')

    return adata, adata_hvg

def preprocessing_atac(adata: AnnData,
                       data_type: str = None,
                       genome: str = None,
                       use_gene_weigt: bool = True,
                       use_top_pcs: bool = False,
                       used_binarize: bool = False,
                       used_hvgs: bool = True,
                       used_lsi_graph: bool = True,
                       min_features: int = 100,
                       min_cells: int = 3,
                       atac_n_top_features=100000,  # or gene list
                       n: int = 15,
                       batch_key: str = 'batch',
                       metric: str = 'euclidean',
                       n_components: int = 50
                       ):
    """
    Preprocess scATAC data matrix.

    Parameters
    ----------
    adata :  AnnData object of shape `n_obs` × `n_vars`. Rows correspond to cells and columns to peaks/regions.
    filter_rate : float, optional
        Proportion for feature selection, by default 0.01
    """
    if min_features is None: min_features = 100
    if atac_n_top_features is None: atac_n_top_features = 100000

    # Preprocessing
    # if type(adata.X) != csr.csr_matrix:
    #     adata.X = scipy.sparse.csr_matrix(adata.X)

    sc.pp.filter_cells(adata, min_genes=min_features)
    sc.pp.filter_genes(adata, min_cells=min_cells)
    sc.pp.normalize_total(adata)
    # adata.raw = adata

    ## TFIDF & LSI
    TFIDF_LSI(adata, n_comps=n_components, binarize=used_binarize, random_state=0)

    if used_lsi_graph:
        # use scanpy functions to do the graph construction
        sc.pp.neighbors(adata, n_neighbors=n, metric=metric, use_rep='X_lsi')

    ## HVP
    adata_hvg = adata.copy()
    if used_hvgs:
        if type(atac_n_top_features) == int and atac_n_top_features > 0 and atac_n_top_features < adata_hvg.shape[1]:
            sc.pp.highly_variable_genes(adata_hvg, n_top_genes=atac_n_top_features, batch_key=batch_key,
                                        inplace=False, subset=True)
        elif type(atac_n_top_features) != int:
            if isinstance(atac_n_top_features, str):
                if os.path.isfile(atac_n_top_features):
                    atac_n_top_features = np.loadtxt(atac_n_top_features, dtype=str)
            adata_hvg = reindex(adata_hvg, atac_n_top_features)

    return adata, adata_hvg

def preprocessing_adt(
        adata: AnnData,
        used_hvgs: bool = False,
        used_pca_graph: bool = True,
        rna_n_top_features=2000,  # or gene list
        n: int = 15,
        batch_key: str = 'batch',
        metric: str = 'euclidean',
        svd_solver: str = 'arpack',
        backed: bool = False
):
    """
    Preprocessing single-cell RNA-seq data

    Parameters
    ----------
    adata
        An AnnData matrice of shape n_obs x n_vars. Rows correspond to cells and columns to genes.
    min_features
        Filtered out cells that are detected in less than n genes. Default: 600.
    min_cells
        Filtered out genes that are detected in less than n cells. Default: 3.
    target_sum
        After normalization, each cell has a total count equal to target_sum. If None, total count of each cell equal to the median of total counts for cells before normalization.
    rna_n_top_features
        Number of highly-variable genes to keep. Default: 2000.

    Return
    -------
    The AnnData object after preprocessing.
    """
    # Preprocessing

    # if not issparse(adata.X):
    if type(adata.X) != csr.csr_matrix and (not backed) and (not adata.isbacked):
        adata.X = scipy.sparse.csr_matrix(adata.X)

    # Normalizing total per cell
    if adata.X.max() < 50:
        print('Warning: adata_adt.X may have already been normalized, do not normalize, please check.')
    else:
        adata = clr_normalize_each_cell(adata)
        adata.layers['norm_data'] = adata.X.copy()

    # Finding variable features for RNA adata
    adata_hvg = adata.copy()
    if used_hvgs:
        if type(rna_n_top_features) == int:
            if rna_n_top_features > len(adata_hvg.var_names):
                rna_n_top_features = len(adata_hvg.var_names)
            if batch_key is not None:
                sc.pp.highly_variable_genes(adata_hvg, n_top_genes=rna_n_top_features, batch_key=batch_key)
            else:
                sc.pp.highly_variable_genes(adata_hvg, n_top_genes=rna_n_top_features)
                # sc.pp.highly_variable_genes(adata, min_mean=0.0125, max_mean=3, min_disp=0.5)
            adata_hvg = adata_hvg[:, adata_hvg.var.highly_variable].copy()
        elif type(rna_n_top_features) != int:
            if isinstance(rna_n_top_features, str):
                if os.path.isfile(rna_n_top_features):
                    rna_n_top_features = np.loadtxt(rna_n_top_features, dtype=str)
            adata_hvg = reindex(adata_hvg, rna_n_top_features)

    if used_pca_graph:
        # scale data, clip values exceeding standard deviation 10.
        if adata_hvg.X.min() < 0:
            print('Warning: adata.X may have already been scaled, do not scale, please check.')
        else:
            sc.pp.scale(adata_hvg, max_value=10)
        # PCA
        sc.tl.pca(adata_hvg, svd_solver=svd_solver)
        # use scanpy functions to do the graph construction
        sc.pp.neighbors(adata_hvg, n_neighbors=n, metric=metric, use_rep='X_pca')
        # 返回norm_data
        adata_hvg.X = adata_hvg.layers['norm_data'].copy()
    else:
        sc.pp.neighbors(adata_hvg, n_neighbors=n, metric=metric, use_rep='X')

    return adata, adata_hvg

### preprocessing main function
def preprocessing(
        adata: [AnnData, MuData],
        profile: str = 'RNA',
        data_type: str = 'Paired',
        sub_data_type: str = ['rna', 'atac'],
        batch_key: str = 'batch',
        weight = 0.8,
        graph_const_method: str = None,
        genome: str = None,
        use_gene_weigt: bool = True,
        use_top_pcs: bool = False,
        min_features: int = 600,
        min_cells: int = 3,
        target_sum: int = None,
        rna_n_top_features=None,  # or gene list
        atac_n_top_features=None,  # or gene list
        svd_components_rna: int = 30,
        svd_components_atac: int = 30,
        cca_components: int = 20,
        cca_max_iter: int = 2000,
        randomized_svd: bool = False,
        filter_prop_initial: int = 0,
        filter_prop_refined: int = 0.3,
        filter_prop_propagated: int = 0,
        n_iters: int = 1,
        svd_runs: int = 1,
        n: int = 15,
        metric: str = 'euclidean',
        svd_solver: str = 'arpack',
        n_components: int = 50,
        keep_mt: bool = False,
        backed: bool = False,
        verbose: bool = True
):
    """
    Preprocessing single-cell data

    Parameters
    ----------
    adata
        An AnnData matrice of shape n_obs x n_vars. Rows correspond to cells and columns to genes.
    profile
        Specify the single-cell profile type, RNA or ATAC, Default: RNA.
    min_features
        Filtered out cells that are detected in less than n genes. Default: 100.
    min_cells
        Filtered out genes that are detected in less than n cells. Default: 3.
    target_sum
        After normalization, each cell has a total count equal to target_sum. If None, total count of each cell equal to the median of total counts for cells before normalization.
    n_top_features
        Number of highly-variable genes to keep. Default: 2000.
    log
        If log, record each operation in the log file. Default: None.

    Return
    -------
    The AnnData object after preprocessing.

    """
    if profile == 'RNA':
        return preprocessing_rna(
            adata,
            min_features=min_features,
            min_cells=min_cells,
            target_sum=target_sum,
            used_hvgs=True,
            used_pca_graph=True,
            rna_n_top_features=rna_n_top_features,
            n=n,
            batch_key=batch_key,
            metric=metric,
            svd_solver=svd_solver,
            keep_mt=keep_mt,
            backed=backed,
        )
    elif profile == 'ATAC':
        return preprocessing_atac(
            adata,
            min_features=min_features,
            min_cells=min_cells,
            used_hvgs=True,
            used_lsi_graph=True,
            atac_n_top_features=atac_n_top_features,
            n=n,
            batch_key=batch_key,
            metric=metric,
            n_components=n_components
        )
    elif profile == 'ADT':
        return preprocessing_adt(
            adata,
            used_hvgs=False,
            used_pca_graph=True,
            rna_n_top_features=rna_n_top_features,  # or gene list
            n=n,
            batch_key=batch_key,
            metric=metric,
            svd_solver=svd_solver,
            backed=backed
        )
    elif profile == 'multi-modal':
        # UnPaired TO DO
        assert data_type in ['Paired', 'UnPaired'], 'Data_type must be "Paired", or "UnPaired".'
        assert sub_data_type in ['rna', 'atac', 'adt'], 'Sub_data_type must be "rna", "atac", or "adt".'

        if len(sub_data_type) == 2:
            if sub_data_type[0] == 'rna' and sub_data_type[1] == 'atac':
                rna_adata = adata.mod['rna']
                atac_adata = adata.mod['atac']
                adt_adata = None
            elif sub_data_type[0] == 'rna' and sub_data_type[1] == 'adt':
                rna_adata = adata.mod['rna']
                atac_adata = None
                adt_adata = adata.mod['adt']
        else:
            ValueError('The length of sub_data_type must be 2, such as: ["rna", "atac"] or ["rna", "adt"].')

        if data_type == 'Paired':
            rna_adata, rna_adata_hvg = preprocessing_rna(
                rna_adata,
                min_features=min_features,
                min_cells=min_cells,
                target_sum=target_sum,
                used_hvgs=True,
                used_pca_graph=True,
                rna_n_top_features=rna_n_top_features,
                n=n,
                batch_key=batch_key,
                metric=metric,
                svd_solver=svd_solver,
                keep_mt=keep_mt,
                backed=backed,
            )
            if atac_adata is not None:
                atac_adata, atac_adata_hvg = preprocessing_atac(
                    atac_adata,
                    data_type,
                    min_features=min_features,
                    min_cells=min_cells,
                    used_hvgs=True,
                    used_lsi_graph=True,
                    atac_n_top_features=atac_n_top_features,
                    n=n,
                    batch_key=batch_key,
                    metric=metric,
                    n_components=n_components
                )
                ## Concatenating different modalities
                adata_paired = ad.concat([rna_adata_hvg, atac_adata_hvg], axis=1)

                ## the .obs layer is empty now, and we need to repopulate it
                rna_cols = rna_adata.obs.columns
                atac_cols = atac_adata.obs.columns

                rnaobs = rna_adata.obs.copy()
                rnaobs.columns = ["rna:" + x for x in rna_cols]
                atacobs = atac_adata.obs.copy()
                atacobs.columns = ["atac:" + x for x in atac_cols]
                adata_paired.obs = pd.merge(rnaobs, atacobs, left_index=True, right_index=True)

                ## 先将 scATAC 转换为基因活性矩阵
                print('Convert peak to gene activity matrix, please wait.', flush=True)
                print('`genome` parameter should be set correctly', flush=True)
                print("Choose from {‘hg19’, ‘hg38’, ‘mm9’, ‘mm10’}", flush=True)
                adata_CG_atac = gene_scores(atac_adata, genome=genome, use_gene_weigt=use_gene_weigt, use_top_pcs=use_top_pcs)

                ## 交集
                common_genes = set(rna_adata.var_names).intersection(set(adata_CG_atac.var_names))
                print('There are {} common genes in RNA and ATAC datasets'.format(len(common_genes)))
                rna_adata_shared = rna_adata[:, list(common_genes)]
                atac_adata_shared = adata_CG_atac[:, list(common_genes)]

                # 通过cell matching 构建组学间的图结构
                print('To start performing cell matching for adjacency matrix of the graph, please wait...', flush=True)
                inter_connect = create_adj(rna_adata,
                                           atac_adata,
                                           rna_adata_shared,
                                           atac_adata_shared,
                                           data_type=data_type,
                                           rna_n_top_features=rna_n_top_features,
                                           atac_n_top_features=atac_n_top_features,
                                           batch_key=batch_key, svd_components1=svd_components_rna,
                                           svd_components2=svd_components_atac,
                                           cca_components=cca_components, cca_max_iter=cca_max_iter,
                                           randomized_svd=randomized_svd,
                                           filter_prop_initial=filter_prop_initial,
                                           filter_prop_refined=filter_prop_refined,
                                           filter_prop_propagated=filter_prop_propagated,
                                           n_iters=n_iters, svd_runs=svd_runs, verbose=verbose)

                # 设置权重
                # Iterate over keys in rna_obsp (assuming atac_obsp has the same keys)
                for key in rna_adata_hvg.obsp.keys():
                    # 计算加权平均的连通性矩阵
                    intra_connect = weight * rna_adata_hvg.obsp[key] + (1 - weight) * atac_adata_hvg.obsp[key]

                adata_paired.obsp['connectivities'] = inter_connect + intra_connect

                return adata_paired

            elif adt_adata is not None:
                adt_adata, adt_adata_hvg = preprocessing_adt(
                    adt_adata,
                    used_hvgs=False,
                    used_pca_graph=True,
                    rna_n_top_features=rna_n_top_features,  # or gene list
                    n=n,
                    batch_key=batch_key,
                    metric=metric,
                    svd_solver=svd_solver,
                    backed=backed
                )
                ## Concatenating different modalities
                adata_paired = ad.concat([rna_adata_hvg, adt_adata_hvg], axis=1)

                ## the .obs layer is empty now, and we need to repopulate it
                rna_cols = rna_adata.obs.columns
                adt_cols = adt_adata.obs.columns

                rnaobs = rna_adata.obs.copy()
                rnaobs.columns = ["rna:" + x for x in rna_cols]
                adtobs = adt_adata.obs.copy()
                adtobs.columns = ["adt:" + x for x in adt_cols]
                adata_paired.obs = pd.merge(rnaobs, adtobs, left_index=True, right_index=True)

                ## 交集
                common_genes = set(rna_adata.var_names).intersection(set(adt_adata.var_names))
                print('There are {} common genes in RNA and ADT datasets'.format(len(common_genes)))
                rna_adata_shared = rna_adata[:, list(common_genes)]
                adt_adata_shared = adt_adata[:, list(common_genes)]

                ## 通过cell matching 构建组学间的图结构
                print('To start performing cell matching for adjacency matrix of the graph, please wait...', flush=True)
                inter_connect = create_adj(rna_adata,
                                           adt_adata,
                                           rna_adata_shared,
                                           adt_adata_shared,
                                           data_type=data_type,
                                           rna_n_top_features=rna_n_top_features,
                                           atac_n_top_features=atac_n_top_features,
                                           batch_key=batch_key, svd_components1=svd_components_rna,
                                           svd_components2=svd_components_atac,
                                           cca_components=cca_components, cca_max_iter=cca_max_iter,
                                           randomized_svd=randomized_svd,
                                           filter_prop_initial=filter_prop_initial,
                                           filter_prop_refined=filter_prop_refined,
                                           filter_prop_propagated=filter_prop_propagated,
                                           n_iters=n_iters, svd_runs=svd_runs, verbose=verbose)

                # Iterate over keys in rna_obsp (assuming adt_obsp has the same keys)
                for key in rna_adata_hvg.obsp.keys():
                    # 计算加权平均的连通性矩阵
                    intra_connect = weight * rna_adata_hvg.obsp[key] + (1 - weight) * adt_adata_hvg.obsp[key]

                adata_paired.obsp['connectivities'] = inter_connect + intra_connect

                return adata_paired

    elif profile == 'spatial':
        assert data_type in ['single-modal', 'multi-modal'], 'Data_type must be "single-modal", or "multi-modal".'

        if data_type == 'single-modal':
            _, rna_adata_hvg = preprocessing_rna(
                adata,
                min_features=min_features,
                min_cells=min_cells,
                target_sum=target_sum,
                used_hvgs=True,
                used_pca_graph=True,
                rna_n_top_features=rna_n_top_features,
                n=n,
                batch_key=batch_key,
                metric=metric,
                svd_solver=svd_solver,
                keep_mt=True,
                backed=backed,
            )
            # Construct Spatial Graph
            if graph_const_method is None:
                graph_dict = graph_construction(adata.obsm['spatial'], adata.shape[0], k=n)
                spatial_adj = graph_dict.toarray()
            elif graph_const_method == 'Squidpy':
                import squidpy as sq
                sq.gr.spatial_neighbors(adata, coord_type="generic", n_neighs=n)
                spatial_adj = adata.obsp['spatial_connectivities'].toarray()

            # Ensure adj is a csr_matrix
            if not isinstance(spatial_adj, csr_matrix):
                spatial_adj = csr_matrix(spatial_adj)

            expr_adj = rna_adata_hvg.obsp['connectivities']
            # Ensure adj is a csr_matrix
            if not isinstance(expr_adj, csr_matrix):
                expr_adj = csr_matrix(expr_adj)

            # 计算加权平均的连通性矩阵
            rna_adata_hvg.obsp['connectivities'] = weight * spatial_adj + (1 - weight) * expr_adj

            return rna_adata_hvg

        elif data_type == 'multi-modal':
            assert sub_data_type[0] in ['rna', 'atac', 'adt'] and sub_data_type[1] in ['rna', 'atac', 'adt'], 'Sub_data_type must be "rna", "atac", or "adt".'

            if len(sub_data_type) == 2:
                if sub_data_type[0] == 'rna' and sub_data_type[1] == 'atac':
                    rna_adata = adata.mod['rna']
                    atac_adata = adata.mod['atac']
                    adt_adata = None
                elif sub_data_type[0] == 'rna' and sub_data_type[1] == 'adt':
                    rna_adata = adata.mod['rna']
                    atac_adata = None
                    adt_adata = adata.mod['adt']
            else:
                ValueError('The length of sub_data_type must be 2, such as: ["rna", "atac"] or ["rna", "adt"].')

            rna_adata, rna_adata_hvg = preprocessing_rna(
                rna_adata,
                min_features=min_features,
                min_cells=min_cells,
                target_sum=target_sum,
                used_hvgs=True,
                used_pca_graph=True,
                rna_n_top_features=rna_n_top_features,
                n=n,
                batch_key=batch_key,
                metric=metric,
                svd_solver=svd_solver,
                keep_mt=True,
                backed=backed,
            )
            if atac_adata is not None:
                atac_adata, atac_adata_hvg = preprocessing_atac(
                    atac_adata,
                    data_type,
                    min_features=min_features,
                    min_cells=min_cells,
                    used_hvgs=True,
                    used_lsi_graph=True,
                    atac_n_top_features=atac_n_top_features,
                    n=n,
                    batch_key=batch_key,
                    metric=metric,
                    n_components=n_components
                )
                ## Concatenating different modalities
                adata_paired = ad.concat([rna_adata_hvg, atac_adata_hvg], axis=1)

                ## the .obs layer is empty now, and we need to repopulate it
                rna_cols = rna_adata.obs.columns
                atac_cols = atac_adata.obs.columns

                rnaobs = rna_adata.obs.copy()
                rnaobs.columns = ["rna:" + x for x in rna_cols]
                atacobs = atac_adata.obs.copy()
                atacobs.columns = ["atac:" + x for x in atac_cols]
                adata_paired.obs = pd.merge(rnaobs, atacobs, left_index=True, right_index=True)

                adata_paired.obsm = rna_adata.obsm

                ## 先将 scATAC 转换为基因活性矩阵
                if len(atac_adata.var_names) > 50000:
                    print('Convert peak to gene activity matrix, please wait.', flush=True)
                    print('`genome` parameter should be set correctly', flush=True)
                    print("Choose from {‘hg19’, ‘hg38’, ‘mm9’, ‘mm10’}", flush=True)
                    adata_CG_atac = gene_scores(atac_adata, genome=genome, use_gene_weigt=use_gene_weigt,
                                                use_top_pcs=use_top_pcs)

                    ## 交集
                    common_genes = set(rna_adata.var_names).intersection(set(adata_CG_atac.var_names))
                    print('There are {} common genes in RNA and ATAC datasets'.format(len(common_genes)))
                    rna_adata_shared = rna_adata[:, list(common_genes)]
                    atac_adata_shared = adata_CG_atac[:, list(common_genes)]
                else:
                    ## 交集
                    common_genes = set(rna_adata.var_names).intersection(set(atac_adata.var_names))
                    print('There are {} common genes in RNA and ATAC datasets'.format(len(common_genes)))
                    rna_adata_shared = rna_adata[:, list(common_genes)]
                    atac_adata_shared = atac_adata[:, list(common_genes)]

                # 通过cell matching 构建组学间的图结构
                print('To start performing cell matching for adjacency matrix of the graph, please wait...', flush=True)
                inter_connect = create_adj(rna_adata,
                                           atac_adata,
                                           rna_adata_shared,
                                           atac_adata_shared,
                                           data_type=data_type,
                                           rna_n_top_features=rna_n_top_features,
                                           atac_n_top_features=atac_n_top_features,
                                           batch_key=batch_key, svd_components1=svd_components_rna,
                                           svd_components2=svd_components_atac,
                                           cca_components=cca_components, cca_max_iter=cca_max_iter,
                                           randomized_svd=randomized_svd,
                                           filter_prop_initial=filter_prop_initial,
                                           filter_prop_refined=filter_prop_refined,
                                           filter_prop_propagated=filter_prop_propagated,
                                           n_iters=n_iters, svd_runs=svd_runs, verbose=verbose)

                # Construct Spatial Graph
                if graph_const_method is None:
                    graph_dict = graph_construction(rna_adata_hvg.obsm['spatial'], rna_adata_hvg.shape[0], k=n)
                    spatial_adj = graph_dict #.toarray()
                elif graph_const_method == 'Squidpy':
                    import squidpy as sq
                    sq.gr.spatial_neighbors(rna_adata_hvg, coord_type="generic", n_neighs=n)
                    spatial_adj = rna_adata_hvg.obsp['spatial_connectivities'] #.toarray()

                # Ensure adj is a csr_matrix
                if not isinstance(spatial_adj, csr_matrix):
                    spatial_adj = csr_matrix(spatial_adj)
                if not isinstance(inter_connect, csr_matrix):
                    inter_connect = csr_matrix(inter_connect)

                # 计算加权平均的连通性矩阵
                adata_paired.obsp['connectivities'] = weight * spatial_adj + (1 - weight) * inter_connect

                return adata_paired

            elif adt_adata is not None:
                adt_adata, adt_adata_hvg = preprocessing_adt(
                    adt_adata,
                    used_hvgs=False,
                    used_pca_graph=True,
                    rna_n_top_features=rna_n_top_features,  # or gene list
                    n=n,
                    batch_key=batch_key,
                    metric=metric,
                    svd_solver=svd_solver,
                    backed=backed
                )
                ## Concatenating different modalities
                adata_paired = ad.concat([rna_adata_hvg, adt_adata_hvg], axis=1)

                ## the .obs layer is empty now, and we need to repopulate it
                rna_cols = rna_adata.obs.columns
                adt_cols = adt_adata.obs.columns

                rnaobs = rna_adata.obs.copy()
                rnaobs.columns = ["rna:" + x for x in rna_cols]
                adtobs = adt_adata.obs.copy()
                adtobs.columns = ["adt:" + x for x in adt_cols]
                adata_paired.obs = pd.merge(rnaobs, adtobs, left_index=True, right_index=True)

                adata_paired.obsm = rna_adata.obsm

                ## 交集
                common_genes = set(rna_adata.var_names).intersection(set(adt_adata.var_names))
                print('There are {} common genes in RNA and ADT datasets'.format(len(common_genes)))
                rna_adata_shared = rna_adata[:, list(common_genes)]
                adt_adata_shared = adt_adata[:, list(common_genes)]

                ## 通过cell matching 构建组学间的图结构
                print('To start performing cell matching for adjacency matrix of the graph, please wait...', flush=True)
                inter_connect = create_adj(rna_adata_hvg,
                                           adt_adata,
                                           rna_adata_shared,
                                           adt_adata_shared,
                                           data_type=data_type,
                                           rna_n_top_features=rna_n_top_features,
                                           atac_n_top_features=atac_n_top_features,
                                           batch_key=batch_key, svd_components1=svd_components_rna,
                                           svd_components2=svd_components_atac,
                                           cca_components=cca_components, cca_max_iter=cca_max_iter,
                                           randomized_svd=randomized_svd,
                                           filter_prop_initial=filter_prop_initial,
                                           filter_prop_refined=filter_prop_refined,
                                           filter_prop_propagated=filter_prop_propagated,
                                           n_iters=n_iters, svd_runs=svd_runs, verbose=verbose)

                # Construct Spatial Graph
                if graph_const_method is None:
                    graph_dict = graph_construction(rna_adata_hvg.obsm['spatial'], rna_adata_hvg.shape[0], k=n)
                    spatial_adj = graph_dict #.toarray()
                elif graph_const_method == 'Squidpy':
                    import squidpy as sq
                    sq.gr.spatial_neighbors(rna_adata_hvg, coord_type="generic", n_neighs=n)
                    spatial_adj = rna_adata_hvg.obsp['spatial_connectivities'] #.toarray()

                # Ensure adj is a csr_matrix
                if not isinstance(spatial_adj, csr_matrix):
                    spatial_adj = csr_matrix(spatial_adj)
                if not isinstance(inter_connect, csr_matrix):
                    inter_connect = csr_matrix(inter_connect)

                # 计算加权平均的连通性矩阵
                adata_paired.obsp['connectivities'] = weight * spatial_adj + (1 - weight) * inter_connect
                return adata_paired
    else:
        raise ValueError("Not support profile: `{}` yet".format(profile))


import random
import numpy as np
import muon as mu
# from muon import MuData
import torch
from sklearn.preprocessing import LabelEncoder
from torch_geometric.data import Data
# from torch_scatter import scatter


def Transfer_scData(adata, label_name=None, data_type=None):
    """
    Converts an AnnData or MuData object into a PyTorch Geometric Data object.

    Parameters
    ----------
    adata : AnnData or MuData
        The annotated data matrix to be converted.
    label_name : str, optional
        Column name in `adata.obs` containing the labels for each node.
        If None, all nodes are assigned a default label of 1.
    data_type : str, optional
        Indicates the type of data matrix ('muData' or 'AnnData').
    Returns
    -------
    torch_geometric.data.Data
        A PyTorch Geometric Data object containing the graph data.
    """
    # Initialize an empty list for edges and their weights
    row_col = []
    # if data_type == 'Paired':
    #     # adj = adata.uns['obsp_rna']['connectivities']
    #     adj = adata.obsp['connectivities_combined']
    # else:
    adj = adata.obsp['connectivities']
    edge_weight = []
    rows, cols = adj.nonzero()
    edge_nums = adj.getnnz()
    for i in range(edge_nums):
        row_col.append([rows[i], cols[i]])
        edge_weight.append(adj.data[i])
    edge_index = torch.tensor(np.array(row_col), dtype=torch.long).T
    edge_attr = torch.tensor(np.array(edge_weight), dtype=torch.float)
    edge_attr = torch.unsqueeze(edge_attr, dim=-1)  # 在第一维度添加维度

    ## celltype label_encoder Todo 如果标签信息本身就是数值则不需要转换
    if label_name is not None:
        label_encoder = LabelEncoder()
        if data_type == 'Paired':
            meta = np.array(adata.obs['rna:' + label_name].astype('str'))
        else:
            meta = np.array(adata.obs[label_name].astype('str'))
        meta = label_encoder.fit_transform(meta)
        # meta = meta.astype(np.float32)
        inverse = label_encoder.inverse_transform(range(0, np.max(meta) + 1))
        y = torch.from_numpy(meta.astype(np.int64))
    else:
        y = torch.from_numpy(np.array([1] * adata.shape[0]))

    if type(adata.X) == np.ndarray:
        data = Data(edge_index=edge_index, edge_attr=edge_attr, x=torch.FloatTensor(adata.X),  # .todense()
                    y=torch.LongTensor(y))
    else:
        data = Data(edge_index=edge_index, edge_attr=edge_attr, x=torch.FloatTensor(adata.X.todense()),  # .todense()
                    y=torch.LongTensor(y))
    return data


def split_data(num_y, val_split, test_split):
    split_idx = list(range(num_y))
    random.shuffle(split_idx)
    train_split = (1 - test_split - val_split)
    train_idx = split_idx[: int(len(split_idx) * train_split)] # Train mask
    train_mask = torch.zeros(num_y, dtype = torch.bool)
    train_mask[train_idx] = 1
    val_idx = split_idx[ int(len(split_idx) * train_split) : int(len(split_idx) * (train_split + val_split))] # Val mask
    val_mask = torch.zeros(num_y, dtype=torch.bool)
    val_mask[val_idx] = 1
    test_idx = split_idx[int(len(split_idx) * (train_split + val_split)) :] # Test mask
    test_mask = torch.zeros(num_y, dtype=torch.bool)
    test_mask[test_idx] = 1
    return train_mask, val_mask, test_mask


from torch_geometric.data import Data
from torch_geometric.loader import NeighborLoader, GraphSAINTEdgeSampler
from torch_geometric.utils import structured_negative_sampling
def generate_batch(data, edge_attr_dict, mask, batch_size, loader_type="graphsaint",
                   num_layers=2):
    # Positive edges
    if mask == "train":
        pos_edge_index = data.edge_index[:, data.train_mask]
        edge_type = data.edge_attr[data.train_mask]
        y = data.y[data.train_mask]
    elif mask == "val":
        pos_edge_index = data.edge_index[:, data.val_mask]
        edge_type = data.edge_attr[data.val_mask]
        y = data.y[data.train_mask]
    elif mask == "test":
        pos_edge_index = data.edge_index[:, data.test_mask]
        edge_type = data.edge_attr[data.test_mask]
        y = data.y[data.train_mask]
    else:
        pos_edge_index = data.edge_index
        edge_type = data.edge_attr
        y = data.y[data.train_mask]

    # Negative edges
    neg_edge_index, neg_edge_type = negative_sampler(pos_edge_index, edge_type, edge_attr_dict)

    # All edges and labels
    total_edge_index = torch.cat([pos_edge_index, neg_edge_index], dim=-1)
    total_edge_type = torch.cat([edge_type, neg_edge_type], dim=-1)

    # Save information for the loader
    data = Data(x=data.x, edge_index=total_edge_index, edge_attr=total_edge_type, y=y)
    data.n_id = torch.arange(data.num_nodes)
    if loader_type == "neighbor":
        loader = NeighborLoader(data, num_neighbors=[-1] * num_layers, batch_size=batch_size,
                                input_nodes=torch.arange(data.num_nodes), shuffle=True)
    elif loader_type == "graphsaint":
        # loader = GraphSAINTRandomWalkSampler(data, batch_size = batch_size, walk_length = num_layers)
        loader = GraphSAINTEdgeSampler(data, batch_size=batch_size, num_steps=16)
    else:
        raise NotImplementedError

    return loader


def negative_sampler(pos_edge_index, edge_type, edge_attr_dict):
    if len(edge_type) == 0: return pos_edge_index, edge_type
    neg_edge_index = None
    neg_edge_type = []
    for attr, idx in edge_attr_dict.items():
        mask = (edge_type == idx)
        if mask.sum() == 0: continue
        pos_rel_edge_index = pos_edge_index.T[mask].T
        neg_source, neg_target, neg_rand = structured_negative_sampling(pos_rel_edge_index)
        neg_rel_edge_index = torch.stack((neg_source, neg_rand), dim=0)
        """
        neg_rel_edge_index = pos_rel_edge_index.clone()
        rand_axis = random.sample([0, 1], 1)[0]
        rand_index = torch.randperm(pos_rel_edge_index.size(1))
        neg_rel_edge_index[rand_axis, :] = pos_rel_edge_index[rand_axis, rand_index]
        """
        if neg_edge_index == None:
            neg_edge_index = neg_rel_edge_index
        else:
            neg_edge_index = torch.cat((neg_edge_index, neg_rel_edge_index), 1)

        neg_edge_type.extend([idx] * mask.sum())

    return neg_edge_index, torch.tensor(neg_edge_type)


def to_dense_adj(edge_index, edge_attr, num_nodes, device, alpha=2):
    # edge_index: [2, E] tensor of edge indices
    # edge_attr: [E, D] tensor of edge attributes
    # num_nodes: number of nodes in the graph

    # 初始化一个邻接矩阵，大小为[num_nodes, num_nodes]
    adj_matrix = alpha * torch.ones(num_nodes, num_nodes).to(device)

    # 使用scatter函数将边属性加到邻接矩阵的对应位置
    # edge_index[0] 是起始节点，edge_index[1] 是终止节点
    adj_matrix[edge_index[0], edge_index[1]] = edge_attr/alpha

    return adj_matrix


def get_prior(celltype1, celltype2, alpha=2):
    """
    Create a prior correspondence matrix according to cell labels

    Parameters
    ----------
    celltype1
        cell labels of dataset X
    celltype2
        cell labels of dataset Y
    alpha
        the confidence of label, ranges from (1, inf). Higher alpha means better confidence. Default: 2.0

    Return
    ------
    torch.tensor
        a prior correspondence matrix between cells
    """

    Couple = alpha * torch.ones(len(celltype1), len(celltype2))

    for i in set(celltype1):
        index1 = np.where(celltype1 == i)
        if i in set(celltype2):
            index2 = np.where(celltype2 == i)
            for j in index1[0]:
                Couple[j, index2[0]] = 1 / alpha

    return Couple

def scipy_sparse_mat_to_torch_sparse_tensor(sparse_mx):
    sparse_mx = sparse_mx.tocoo().astype(np.float32)
    indices = torch.from_numpy(
        np.vstack((sparse_mx.row, sparse_mx.col)).astype(np.int64))
    values = torch.from_numpy(sparse_mx.data)
    shape = torch.Size(sparse_mx.shape)
    return torch.sparse_coo_tensor(indices, values, shape) # torch.sparse.FloatTensor(indices, values, shape)


def extract_subgraph(data, label):
    # 找出所有具有特定标签的节点索引
    node_indices = (data.y == label).nonzero(as_tuple=True)[0]

    # 提取这些节点的特征
    sub_x = data.x[node_indices]
    sub_y = data.y[node_indices]

    # 提取包含这些节点的边
    edge_index = data.edge_index
    mask = torch.isin(edge_index[0], node_indices) & torch.isin(edge_index[1], node_indices)
    sub_edge_index = edge_index[:, mask]

    # 将边索引调整为新的节点索引
    _, new_indices = node_indices.unique(return_inverse=True)
    remap = {old_idx.item(): new_idx for old_idx, new_idx in zip(node_indices, new_indices)}
    for i in range(sub_edge_index.shape[1]):
        sub_edge_index[0, i] = remap[sub_edge_index[0, i].item()]
        sub_edge_index[1, i] = remap[sub_edge_index[1, i].item()]

    # 提取对应的边特征
    sub_edge_attr = data.edge_attr[mask] if data.edge_attr is not None else None

    return Data(x=sub_x, edge_index=sub_edge_index, y=sub_y, edge_attr=sub_edge_attr)


import io
import logging
import pickle

import numpy as np
import torch
from anndata import AnnData
from scipy.sparse import csr_matrix, hstack

logger = logging.getLogger(__name__)


def _validate_var_names(adata, source_var_names):
    # Warning for gene percentage
    user_var_names = adata.var_names
    try:
        percentage = (len(user_var_names.intersection(source_var_names)) / len(user_var_names)) * 100
        percentage = round(percentage, 4)
        if percentage != 100:
            logger.warning(f"WARNING: Query shares {percentage}% of its genes with the reference."
                           "This may lead to inaccuracy in the results.")
    except Exception:
        logger.warning("WARNING: Something is wrong with the reference genes.")

    user_var_names = user_var_names.astype(str)
    new_adata = adata

    # Get genes in reference that are not in query
    ref_genes_not_in_query = []
    for name in source_var_names:
        if name not in user_var_names:
            ref_genes_not_in_query.append(name)

    if len(ref_genes_not_in_query) > 0:
        print("Query data is missing expression data of ",
              len(ref_genes_not_in_query),
              " genes which were contained in the reference dataset.")
        print("The missing information will be filled with zeroes.")

        filling_X = np.zeros((len(adata), len(ref_genes_not_in_query)))
        if isinstance(adata.X, csr_matrix):
            filling_X = csr_matrix(filling_X)  # support csr sparse matrix
            new_target_X = hstack((adata.X, filling_X))
        else:
            new_target_X = np.concatenate((adata.X, filling_X), axis=1)
        new_target_vars = adata.var_names.tolist() + ref_genes_not_in_query
        new_adata = AnnData(new_target_X, dtype="float32")
        new_adata.var_names = new_target_vars
        new_adata.obs = adata.obs.copy()

    if len(user_var_names) - (len(source_var_names) - len(ref_genes_not_in_query)) > 0:
        print(
            "Query data contains expression data of ",
            len(user_var_names) - (len(source_var_names) - len(ref_genes_not_in_query)),
            " genes that were not contained in the reference dataset. This information "
            "will be removed from the query data object for further processing.")

        # remove unseen gene information and order anndata
        new_adata = new_adata[:, source_var_names].copy()

    print(new_adata)

    return new_adata


class UnpicklerCpu(pickle.Unpickler):
    """Helps to pickle.load a model trained on GPU to CPU.

    See also https://github.com/pytorch/pytorch/issues/16797#issuecomment-633423219.
    """

    def find_class(self, module, name):
        if module == 'torch.storage' and name == '_load_from_bytes':
            return lambda b: torch.load(io.BytesIO(b), map_location='cpu')
        else:
            return super().find_class(module, name)


# plot utils
## adapted from https://github.com/theislab/scarches/blob/51a0294ca987dabffb6d109178e0f69a90f9c24f/scarches/plotting/_alluvial.py#L35
import numpy as np
from collections import Counter, defaultdict
import matplotlib.pyplot as plt
from matplotlib.patches import Polygon
import matplotlib.cm
import itertools
# import bidi.algorithm  # for RTL languages


def plot_alluvial(input_data, *args, **kwargs):
    at = AlluvialTool(input_data, *args, **kwargs)
    ax = at.plot(**kwargs)
    ax.axis('off')
    return ax


class AlluvialTool:
    def __init__(
            self, input_data=(), x_range=(0, 1), res=20, h_gap_frac=0.03, v_gap_frac=0.03, **kwargs):
        self.input = input_data
        self.x_range = x_range
        self.res = res  # defines the resolution of the splines for all veins
        self.combs = sorted(itertools.product((0, 1), (1, 0)), key=lambda xy: all(xy))
        self.trace_xy = self.make_vein_blueprint_xy_arrays()
        self.data_dic = self.read_input()
        self.item_widths_dic = self.get_item_widths_dic()
        self.a_members, self.b_members = self.get_item_groups(**kwargs)
        self.h_gap_frac = h_gap_frac
        self.h_gap = x_range[1] * h_gap_frac
        self.v_gap_frac = v_gap_frac
        self.v_gap = sum(
            [width for b_item_counter in self.data_dic.values() for width in b_item_counter.values()]
        ) * v_gap_frac
        self.group_widths = self.get_group_widths()
        self.item_coord_dic = self.make_item_coordinate_dic()
        self.alluvial_fan = self.generate_alluvial_fan()
        self.item_text_len, self.width_text_len = self.get_vein_label_lengths()

    def make_vein_blueprint_xy_arrays(self):
        y = np.array([0, 0.15, 0.5, 0.85, 1])
        x = np.linspace(self.x_range[0], self.x_range[-1], len(y))
        z = np.polyfit(x, y, 4)
        f = np.poly1d(z)

        blueprint_x_vals = np.linspace(x[0], x[-1], self.res)
        blueprint_y_vals = f(blueprint_x_vals)
        return blueprint_x_vals, blueprint_y_vals

    def get_vein_polygon_xy(self, y_range, width):
        x, y = self.trace_xy
        y0, yn = y_range
        scale = yn - y0
        ty = y * scale + y0
        x_new = np.concatenate([x, x[::-1], ])
        y_new = np.concatenate([ty, ty[::-1] + width, ])
        return np.array([x_new, y_new]).transpose()

    def read_input_from_list(self):
        data_table = np.array(self.input)
        data_dic = defaultdict(Counter)
        for line in data_table:
            data_dic[line[0]][line[1]] += 1
        return data_dic

    def read_input_from_dict(self):
        # data_dic = self.input
        # data_table = []
        # for x_item, y_item_counter in data_dic.items():
        #     for y_item, count in y_item_counter.items():
        #         data_table += [[x_item, y_item]] * count
        # data_table = np.array(sorted(data_table))
        # return data_table, data_dic
        return self.input

    def read_input(self):
        return {
            dict: self.read_input_from_dict(),
            list: self.read_input_from_list()
         }[type(self.input)]

    def get_item_widths_dic(self):
        iwd = Counter()
        for a_item, b_item_counter in self.data_dic.items():
            for b_item, width in b_item_counter.items():
                iwd[a_item] += width
                iwd[b_item] += width
        return iwd

    def get_item_groups(self, a_sort=None, b_sort=None, **kwargs):
        _ = kwargs
        a_members = sorted(
            {a_item for a_item in self.data_dic}, key=lambda x: self.item_widths_dic[x]
        ) if not a_sort else a_sort
        b_members = sorted(
            {b_item for b_item_counter in self.data_dic.values() for b_item in b_item_counter},
            key=lambda x: self.item_widths_dic[x]
        ) if not b_sort else b_sort
        return a_members, b_members

    def get_group_widths(self):
        return [self.get_group_width(group) for group in (self.a_members, self.b_members)]

    def make_item_coordinate_dic(self, ):
        item_coord_dic = defaultdict(ItemCoordRecord)
        groups = self.a_members, self.b_members
        group_widths = self.group_widths
        for ind, group in enumerate(groups):
            last_pos = (max(group_widths) - group_widths[ind]) / 2
            for item in group:
                width = self.item_widths_dic[item]
                xy = (self.x_range[ind], last_pos)
                item_coord_dic[item].set_start_state(width, xy, side=ind)
                last_pos += width + self.v_gap
        return item_coord_dic

    def get_group_width(self, group):
        return sum([self.item_widths_dic[item] for item in group]) + (len(group) - 1) * self.v_gap

    def generate_alluvial_vein(self, a_item, b_item):
        width = self.data_dic[a_item][b_item]
        a_item_coord = self.item_coord_dic[a_item].read_state_and_advance_y(width)
        b_item_coord = self.item_coord_dic[b_item].read_state_and_advance_y(width)
        y_range = (a_item_coord[1], b_item_coord[1],)
        return self.get_vein_polygon_xy(y_range, width)

    def get_label_rectangles_xy(self, a_item, b_item):
        width = self.data_dic[a_item][b_item]
        return (
            self.generate_item_sub_rectangle(a_item, width),
            self.generate_item_sub_rectangle(b_item, width),
        )

    def generate_item_sub_rectangle(self, item, width):
        dic_entry = self.item_coord_dic[item]
        item_coord = dic_entry.read_state()
        sign = dic_entry.get_side_sign()
        return self.get_rectangle_xy(item_coord, width, sign)

    def get_rectangle_xy(self, item_coord, width, sign):
        x, y = item_coord
        rect = [[
                    x + sign * 0.5 * (0.5 + xa) * self.h_gap,
                    y + ya * width,
                ] for xa, ya in self.combs]
        return np.array(rect)

    def generate_alluvial_fan(self, ):
        alluvial_fan = []
        for a_item in self.a_members:
            b_items4a_item = self.data_dic[a_item].keys()
            for b_item in self.b_members:
                if b_item in b_items4a_item:
                    l_a_rect, l_b_rect = self.get_label_rectangles_xy(a_item, b_item)
                    alluvial_fan += [
                        [self.generate_alluvial_vein(a_item, b_item), l_a_rect, l_b_rect, a_item, b_item, ]]
        return np.array(alluvial_fan)

    def plot(self, figsize=(10, 15), alpha=0.5, **kwargs):
        colors = self.get_color_array(**kwargs)
        fig, ax = plt.subplots(figsize=figsize)
        for num in (0, 1, 2):
            patches = [
                Polygon(item, facecolor=colors[ind], alpha=alpha,
                        ) for ind, item in enumerate(self.alluvial_fan[:, num])
                ]
            for patch in patches:
                ax.add_patch(patch)
        self.auto_label_veins(**kwargs)
        self.label_sides(**kwargs)
        ax.autoscale()
        return ax

    def get_color_array(self, colors=None, color_side=0, rand_seed=1, cmap=None, **kwargs):
        _ = kwargs
        color_items = self.b_members if color_side else self.a_members
        lci = len(color_items)
        if rand_seed is not None:
            np.random.seed(rand_seed)
        cmap = cmap if cmap is not None else matplotlib.cm.get_cmap('hsv', lci * 10 ** 3)
        color_array = colors if colors is not None else [
            cmap(item) for ind, item in enumerate(np.random.rand(lci))]
        ind_dic = {item: ind for ind, item in enumerate(color_items)}
        polygon_colors = []
        for _, _, _, a_item, b_item, in self.alluvial_fan:
            item = b_item if color_side else a_item
            polygon_colors += [color_array[ind_dic[item]]]
        return np.array(polygon_colors)

    def get_vein_label_lengths(self):
        item_text_len = max([len(it) for it in self.item_widths_dic])
        width_text_len = max([len(str(w)) for w in self.item_widths_dic.values()])
        return item_text_len, width_text_len

    def auto_label_veins(self, fontname='Arial', **kwargs):
        # shift = max([len(item) for item in self.item_coord_dic.keys()]) / 50
        for item, vein in self.item_coord_dic.items():
            y_width = vein.get_width()
            sign = vein.get_side_sign()
            side = int(sign + 1) // 2
            ha = 'left' if side else 'right'
            plt.text(
                vein.get_x() + 1.5 * sign * self.h_gap,
                vein.get_y() + y_width / 2,
                self.item_text(item, side, **kwargs),
                ha=ha, va='center', fontname=fontname)

    def label_sides(
            self, labels=None, label_shift=0, disp_width=False, wdisp_sep=7*' ', fontname='Arial', **kwargs):
        if labels is not None:
            _ = kwargs
            y = max(self.group_widths)/2
            itl, wtl = self.item_text_len, self.width_text_len
            for side, sign in enumerate((-1, 1)):
                plt.text(
                    self.x_range[side]+sign*(label_shift+itl+int(disp_width)*(len(wdisp_sep)+wtl))*self.h_gap_frac,
                    y,
                    labels[side],
                    # bidi.algorithm.get_display(labels[side]),  # RTL languages
                    ha='center', va='center', fontname=fontname, fontsize=13, rotation=90-180*side
                )

    def item_text(
            self, item, side,
            disp_width=False, wdisp_sep=7*' ', width_in=True, **kwargs):
        _ = kwargs
        f_item = item
        # f_item = bidi.algorithm.get_display(item)  # for RTL languages
        tal = '<' if f_item == item else '>'
        if not disp_width:
            ans = ('{:%s}' % tal).format(item)
        else:
            width = self.item_coord_dic[item].get_width()
            if side and width_in or (not side and not width_in):
                lc, rc, wl, wr, tl, tr = '>', tal, self.width_text_len, self.item_text_len, width, f_item,
            else:
                lc, rc, wl, wr, tl, tr = tal, '>', self.item_text_len, self.width_text_len, f_item, width,
            pat = '{:%s%d}%s{:%s%d}' % (lc, wl, wdisp_sep, rc, wr,)
            ans = pat.format(tl, tr, )
        return ans


class ItemCoordRecord:
    def __init__(self, ):
        self.width = 0
        self.xy = ()
        self.curr_xy = self.xy[:]
        self.side = -1

    def set_start_state(self, width, xy, side):
        self.width = width
        self.xy = xy
        self.curr_xy = list(self.xy[:])
        self.side = side

    def read_state_and_advance_y(self, width):
        out = self.curr_xy[:]
        self.curr_xy[1] += width
        return out

    def read_state_and_advance_x(self, width):
        out = self.curr_xy[:]
        self.curr_xy[0] += width
        return out

    def read_state(self):
        return self.curr_xy[:]

    def get_xy(self, ):
        return self.xy

    def get_x(self, ):
        return self.xy[0]

    def get_y(self, ):
        return self.xy[1]

    def get_width(self, ):
        return self.width

    def get_side_sign(self, ):
        return 1 if self.side else -1


"""Principal component analysis"""
import numpy as np
from kneed import KneeLocator

def locate_elbow(x, y, S=10, min_elbow=0,
                 curve='convex', direction='decreasing', online=False,
                 **kwargs):
    """Detect knee points

    Parameters
    ----------
    x : `array_like`
        x values
    y : `array_like`
        y values
    S : `float`, optional (default: 10)
        Sensitivity
    min_elbow: `int`, optional (default: 0)
        The minimum elbow location
    curve: `str`, optional (default: 'convex')
        Choose from {'convex','concave'}
        If 'concave', algorithm will detect knees,
        If 'convex', algorithm will detect elbows.
    direction: `str`, optional (default: 'decreasing')
        Choose from {'decreasing','increasing'}
    online: `bool`, optional (default: False)
        kneed will correct old knee points if True,
        kneed will return first knee if False.
    **kwargs: `dict`, optional
        Extra arguments to KneeLocator.

    Returns
    -------
    elbow: `int`
        elbow point
    """
    kneedle = KneeLocator(x[int(min_elbow):], y[int(min_elbow):],
                          S=S, curve=curve,
                          direction=direction,
                          online=online,
                          **kwargs,
                          )
    if(kneedle.elbow is None):
        elbow = len(y)
    else:
        elbow = int(kneedle.elbow)
    return elbow

def select_pcs(adata,
               n_pcs=None,
               S=1,
               curve='convex',
               direction='decreasing',
               online=False,
               min_elbow=None,
               **kwargs):
    """select top PCs based on variance_ratio

    Parameters
    ----------
    n_pcs: `int`, optional (default: None)
        If n_pcs is None,
        the number of PCs will be automatically selected with "`kneed
        <https://kneed.readthedocs.io/>`__"
    S : `float`, optional (default: 1)
        Sensitivity
    min_elbow: `int`, optional (default: None)
        The minimum elbow location
        By default, it is n_components/10
    curve: `str`, optional (default: 'convex')
        Choose from {'convex','concave'}
        If 'concave', algorithm will detect knees,
        If 'convex', algorithm will detect elbows.
    direction: `str`, optional (default: 'decreasing')
        Choose from {'decreasing','increasing'}
    online: `bool`, optional (default: False)
        kneed will correct old knee points if True,
        kneed will return first knee if False.
    **kwargs: `dict`, optional
        Extra arguments to KneeLocator.
    Returns

    """
    if(n_pcs is None):
        n_components = adata.obsm['X_pca'].shape[1]
        if(min_elbow is None):
            min_elbow = n_components/10
        n_pcs = locate_elbow(range(n_components),
                             adata.uns['pca']['variance_ratio'],
                             S=S,
                             curve=curve,
                             min_elbow=min_elbow,
                             direction=direction,
                             online=online,
                             **kwargs)
        adata.uns['pca']['n_pcs'] = n_pcs
    else:
        adata.uns['pca']['n_pcs'] = n_pcs




import warnings
from types import SimpleNamespace

import torch
from ._layers import GATEncoder, GCNEncoder, GNNModelVAE

class Garfield(torch.nn.Module):
    """
    Garfield: Graph-based Contrastive Learning enable Fast Single-Cell Embedding
    """
    def __init__(self, number_of_genes, hidden_dims, bottle_neck_neurons, num_classes,
                 num_heads, dropout, concat, svd_q, used_edge_weight, used_DSBN,
                 used_recon_exp, conv_type, cluster_num, gnn_layer):
        """
        :param args: Arguments object.
        :param number_of_genes: Number of feature of each node.
        :param num_classes: Number of batch.
        """
        super(Garfield, self).__init__()
        self.num_features = number_of_genes
        self.hidden_dims = hidden_dims
        self.bottle_neck_neurons = bottle_neck_neurons
        self.num_heads = num_heads
        self.dropout = dropout
        self.concat = concat
        self.svd_q = svd_q
        self.used_edge_weight = used_edge_weight
        self.used_DSBN = used_DSBN
        self.n_domain = num_classes
        self.used_recon_exp = used_recon_exp
        self.conv_type = conv_type
        self.cluster_num = cluster_num
        self.gnn_layer = gnn_layer
        self.freeze = False
        self.setup_layers()

    def setup_layers(self):
        """
        Creating the layers.
        """
        ## 选择 encoder
        ## 设定参数
        if self.conv_type in ['GAT', 'GATv2Conv']:
            encoder = GATEncoder(
                in_channels=self.num_features,
                hidden_dims=self.hidden_dims,
                latent_dim=self.bottle_neck_neurons,
                conv_type=self.conv_type,
                num_heads=self.num_heads,
                dropout=self.dropout,
                concat=self.concat,
                num_domains=1,  # batch normalization
                svd_q=self.svd_q,
                used_edge_weight=self.used_edge_weight,
                used_DSBN=self.used_DSBN
            )
        else:
            encoder = GCNEncoder(
                in_channels=self.num_features,
                hidden_channels=self.hidden_dims,
                out_channels=self.bottle_neck_neurons,
                dropout=self.dropout,
                num_domains=1,  # batch normalization
                svd_q=self.svd_q,
                used_edge_weight=self.used_edge_weight,
                used_DSBN=self.used_DSBN
            )

        ## GCNModelVAE
        self.VGAE = GNNModelVAE(
            encoder,
            self.bottle_neck_neurons,
            self.hidden_dims,
            self.num_features,
            self.num_heads,
            self.dropout,
            self.concat,
            self.n_domain,
            self.used_edge_weight,
            self.used_DSBN,
            self.used_recon_exp,
            self.conv_type,
            self.cluster_num,
            self.gnn_layer
        )

    def load_model(self, path):
        """
        Load trained model parameters dictionary.
        Parameters
        ----------
        path
            file path that stores the model parameters
        """
        pretrained_dict = torch.load(path, map_location=lambda storage, loc: storage)
        model_dict = self.state_dict()
        pretrained_dict = {k: v for k, v in pretrained_dict.items() if k in model_dict}
        model_dict.update(pretrained_dict)
        self.load_state_dict(model_dict)

    ## embedding
    def encodeBatch(self, data):
        if self.used_recon_exp:
            return self.VGAE.encodeBatch(data)  # mean, recon_features
        else:
            return self.VGAE.encodeBatch(data)  # mean

    def forward(self, data):
        """
        Forward pass with graphs.
        :param data: Data dictionary.
        """
        if self.used_recon_exp:
            z, z_1, z_2, c_1, c_2, mu, logstd, recon_features = self.VGAE(data)
            return z, z_1, z_2, c_1, c_2, mu, logstd, recon_features
        else:
            z, z_1, z_2, c_1, c_2, mu, logstd = self.VGAE(data)
            return z, z_1, z_2, c_1, c_2, mu, logstd